import pytest

import astropy.units as u

import roentgen
from roentgen.lines import get_lines

# remove H and He
all_elements = list(roentgen.elements["symbol"])[2:]


@pytest.mark.parametrize("element_str", all_elements)
def test_line(element_str):
    """Check that all elements return at least one line"""
    assert len(get_lines(0 * u.keV, 100 * u.keV, element=element_str)) > 0


def test_get_lines():
    """Check that lines are returned in a range"""
    assert len(get_lines(4 * u.keV, 6 * u.keV)) > 0


def test_get_all_lines():
    """Check that all lines are returned if the range is large enough."""
    assert len(get_lines(0 * u.keV, 200 * u.keV)) == len(roentgen.emission_lines)


@pytest.mark.parametrize(
    "energy_low,energy_high,element,result",
    [
        (0 * u.eV, 520 * u.eV, None, 15),
        (74.9 * u.keV, 81.0 * u.keV, None, 11),
        (6 * u.keV, 8 * u.keV, "Fe", 3),
        (8 * u.keV, 8.1 * u.keV, "Cu", 2),
    ],
)
def test_get_right_number_of_lines(energy_low, energy_high, element, result):
    """Check a few specific cases for known number of lines"""
    assert len(get_lines(energy_low, energy_high, element=element)) == result



import numpy as np

import astropy.units as u
from astropy.table import Table

import roentgen
from roentgen.util import get_atomic_number

__all__ = ["get_lines"]


@u.quantity_input(energy_low=u.keV, energy_high=u.keV, equivalencies=u.spectral())
def get_lines(energy_low, energy_high, element=None):
    """
    Retrieve all emission lines in an energy range.

    Parameters
    ----------
    energy_low : `astropy.units.Quantity`
        The low end of the energy range

    energy_high : `astropy.units.Quantity`
        The high end of the energy range

    element : str, optional
        Select only lines from a specific element

    Returns
    -------
    line_list : `astropy.table.QTable`
    """
    result = Table()  # this is the default result

    em = roentgen.emission_lines
    energies = em[em.colnames[0]]
    bool_array = (energies < energy_high) * (energies > energy_low)
    if np.any(bool_array):
        result = em.loc[energy_low.to("eV").value : energy_high.to("eV").value]

    if len(result) > 1 and element is not None:
        # check to see if any lines from selected element exist in energy range
        if np.any(result["z"] == get_atomic_number(element)):
            result = result.loc["z", get_atomic_number(element)]

    return result



"""
Use the ``bokeh serve`` command to run the example by executing:
    bokeh serve --show gui
in your browser.
"""
from os.path import dirname, join
import numpy as np

from bokeh.io import curdoc
from bokeh.layouts import layout, Spacer
from bokeh.models import ColumnDataSource, CustomJS
from bokeh.models import HoverTool
from bokeh.models.widgets import (
    Slider,
    TextInput,
    Select,
    Paragraph,
    AutocompleteInput,
    CheckboxGroup,
    TableColumn,
    DataTable,
    Button,
    CheckboxGroup,
    Toggle,
    Select,
    Div
)
from bokeh.plotting import figure
from bokeh.events import ButtonClick

import astropy.units as u
from astropy import constants as const
from astropy.units.imperial import deg_F, inch, foot, mil

from roentgen.absorption import Material, Response
from roentgen.util import get_material_density, density_ideal_gas
import roentgen

u.imperial.enable()

DEFAULT_MATERIAL = ["silicon"]
DEFAULT_THICKNESS = [100.0]
DEFAULT_ENERGY_LOW = 1.0
DEFAULT_ENERGY_HIGH = 200.0
DEFAULT_DENSITY = [get_material_density(DEFAULT_MATERIAL[0]).value]
NUMBER_OF_MATERIALS = len(DEFAULT_MATERIAL)

DEFAULT_DETECTOR_MATERIAL = 'cdte'
DEFAULT_DETECTOR_THICKNESS = 1
DEFAULT_DETECTOR_DENSITY = get_material_density(DEFAULT_DETECTOR_MATERIAL).value

DEFAULT_ENERGY_RESOLUTION = 0.25

DEFAULT_AIR_THICKNESS = 1
DEFAULT_AIR_PRESSURE = 1
DEFAULT_AIR_TEMPERATURE = 25

PLOT_HEIGHT = 400
PLOT_WIDTH = 900
TOOLS = "pan,wheel_zoom,box_zoom,box_select,undo,redo,save,reset"

custom_hover = HoverTool(
    tooltips=[
        ('energy [keV]',   '@{x}{0.2f}'),
        ('transmission',  '@{y}{0.3f}'),  # use @{ } for field names with spaces
    ],
    # display a tooltip whenever the cursor is vertically in line with a glyph
    mode='vline'
)

# defaults
material_list = []

this_material = Material(DEFAULT_MATERIAL[0], DEFAULT_THICKNESS[0] * u.micron)
air_density = density_ideal_gas(DEFAULT_AIR_PRESSURE * const.atm,
                                DEFAULT_AIR_TEMPERATURE * u.Celsius)
air = Material('air', DEFAULT_AIR_THICKNESS * u.m, density=air_density)
this_detector = Material(DEFAULT_DETECTOR_MATERIAL,
                         DEFAULT_DETECTOR_THICKNESS * u.mm)

response = Response(optical_path=[this_material, air], detector=this_detector)

energy = u.Quantity(np.arange(DEFAULT_ENERGY_LOW, DEFAULT_ENERGY_HIGH,
                              DEFAULT_ENERGY_RESOLUTION), "keV")

x = energy.value
y = response.response(energy)
source = ColumnDataSource(data={"x": x, "y": y})

all_materials = list(roentgen.elements["name"]) + \
                list(roentgen.compounds["symbol"])
all_materials.sort()
all_materials = [this_material.lower() for this_material in all_materials]

# Set up the plot
plot = figure(
    plot_height=PLOT_HEIGHT,
    plot_width=PLOT_WIDTH,
    tools=TOOLS,
    x_range=[1, 50],
    y_range=[0, 1],
)
plot.yaxis.axis_label = "Transmission fraction"
plot.xaxis.axis_label = "Energy [keV]"
plot.line("x", "y", source=source, line_width=3, line_alpha=0.6)
plot.title.text = f"{response}"
plot.add_tools(custom_hover)
# Set up the inputs
ylog_checkbox = CheckboxGroup(labels=["y-log"], active=[0])

# materials in the path
material_input = AutocompleteInput(title="Material (lowercase)", value=DEFAULT_MATERIAL[0])
material_input.completions = all_materials
material_thickness_input = TextInput(title="thickness", value=str(DEFAULT_THICKNESS[0]))
material_density_input = TextInput(title="density", value=str(this_material.density.value))

air_thickness_input = TextInput(title="air path length", value=str(DEFAULT_AIR_THICKNESS))
air_pressure_input = TextInput(title='air pressure', value=str(DEFAULT_AIR_PRESSURE))
air_temperature_input = TextInput(title='air temperature', value=str(DEFAULT_AIR_TEMPERATURE))

detector_material_input = AutocompleteInput(title='Detector', value=DEFAULT_DETECTOR_MATERIAL)
detector_material_input.completions = all_materials
detector_thickness_input = TextInput(title='thickness', value=str(DEFAULT_DETECTOR_THICKNESS))
detector_density_input = TextInput(title='density', value=str(DEFAULT_DETECTOR_DENSITY))

p = Paragraph(text="", width=500)
p.text = f"Running roentgen version {roentgen.__version__}"

columns = [
    TableColumn(field="x", title="energy [keV]"),
    TableColumn(field="y", title="Percent"),
]
data_table = DataTable(source=source, columns=columns, width=400, height=700)

download_button = Button(label="Download", button_type="success")
download_button.js_on_event(ButtonClick, CustomJS(args=dict(source=source),
                                                  code=open(join(dirname(__file__), "download.js")).read()))


def convert_air_pressure(value, current_unit, new_unit):
    if current_unit == "atm":
        air_pressure = u.Quantity(value * const.atm, "Pa")
    elif current_unit == "torr":
        air_pressure = u.Quantity(value * const.atm / 760., "Pa")
    else:
        air_pressure = u.Quantity(value, current_unit)

    if new_unit == "atm":
        return (air_pressure.to("Pa") / const.atm).value
    elif new_unit == "torr":
        return (air_pressure.to("Pa") / const.atm).value * 760.0
    else:
        return air_pressure.to(new_unit)


def update_response(attrname, old, new):
    # check whether the input variables have changed and update the response
    global response

    if not material_input.disabled:
        if str(material_input.value).lower() in all_materials:
            this_thickness = u.Quantity(material_thickness_input.value,
                                        material_thick_unit.value)
            this_density = u.Quantity(material_density_input.value,
                                      material_density_unit.value)
            this_material = Material(str(material_input.value).lower(), this_thickness,
                                     density=this_density)
    else:
        # if material not selected, just make a bogus material with no thickness
        this_material = Material('Al', 0 * u.mm)

    if not air_pressure_input.disabled:
        if air_pressure_unit.value == "atm":
            air_pressure = u.Quantity(air_pressure_input.value * const.atm, "Pa")
        elif air_pressure_unit.value == "torr":
            air_pressure = u.Quantity(air_pressure_input.value * const.atm / 760.,
                                      "Pa")
        else:
            air_pressure = u.Quantity(air_pressure_input.value,
                                      air_pressure_unit.value)

        air_path_length = u.Quantity(air_thickness_input.value,
                                     air_thick_unit.value)
        air_temperature = u.Quantity(air_temperature_input.value,
                                     air_temp_unit.value).to("Celsius",
                                                             equivalencies=u.temperature())
        air_density = density_ideal_gas(air_pressure, air_temperature)
        air = Material('air', air_path_length, density=air_density)
    else:
        # if air is not selected than just add bogus air with no thickness
        air = Material('air', 0 * u.mm, density=0 * u.g / u.cm**3)

    if not detector_material_input.disabled:
        if str(detector_material_input.value).lower() in all_materials:
            this_thickness = u.Quantity(detector_thickness_input.value,
                                        detector_thick_unit.value)
            this_density = u.Quantity(detector_density_input.value,
                                      detector_density_unit.value)
            this_detector = Material(str(detector_material_input.value).lower(), this_thickness,
                                     density=this_density)
    else:
        this_detector = None
    response = Response(optical_path=[this_material, air], detector=this_detector)


def update_data(attrname, old, new):

    if plot.x_range.start < DEFAULT_ENERGY_LOW:
        energy = u.Quantity(np.arange(DEFAULT_ENERGY_LOW, plot.x_range.end,
                                  DEFAULT_ENERGY_RESOLUTION), "keV")
    else:
        energy = u.Quantity(np.arange(plot.x_range.start, plot.x_range.end,
                                  DEFAULT_ENERGY_RESOLUTION), "keV")

    y = response.response(energy)

    plot.title.text = f"{response}"

    if plot_checkbox_group.active:
        y = np.log10(response.response(energy))
        plot.y_range.start = -4
        plot.y_range.end = 0
        plot.yaxis.axis_label = 'log(Transmission fraction)'

    else:
        plot.y_range.start = 0
        plot.y_range.end = 1
        plot.yaxis.axis_label = 'Transmission fraction'

    source.data = dict(x=energy, y=y)


def toggle_active(new):
    if 0 in new:
        material_input.disabled = False
        material_thick_unit.disabled = False
        material_density_input.disabled = False
    if 0 not in new:
        material_input.disabled = True
        material_thick_unit.disabled = True
        material_density_input.disabled = True
    if 1 in new:
        air_pressure_input.disabled = False
        air_thickness_input.disabled = False
        air_temperature_input.disabled = False
    if 1 not in new:
        air_pressure_input.disabled = True
        air_thickness_input.disabled = True
        air_temperature_input.disabled = True
    if 2 in new:
        detector_material_input.disabled = False
        detector_thickness_input.disabled = False
        detector_density_input.disabled = False
    if 2 not in new:
        detector_material_input.disabled = True
        detector_thickness_input.disabled = True
        detector_density_input.disabled = True
    return new


checkbox_group = CheckboxGroup(labels=["Material", "Air", "Detector"],
                               active=[0, 1, 2])
checkbox_group.on_click(toggle_active)


def update_button_action():
    update_response("update_plot_button", 0, 0)
    update_data("update", 0, 0)


update_plot_button = Button(label="Update Plot", button_type="success")
update_plot_button.on_click(update_button_action)

plot.x_range.on_change('start', update_data)
plot.x_range.on_change('end', update_data)


plot_checkbox_group = CheckboxGroup(labels=["ylog"], active=[])

length_units = ["m", "mm", "micron", "inch", "foot", "mil"]
pressure_units = ["Pa", "torr", "atm"]
density_units = ["g / cm ** 3", "kg / m ** 3"]
temperature_units = ["K", "deg_C", "deg_F"]


material_thick_unit = Select(title="unit", value=length_units[2], options=length_units)


def update_mat_thick_units(attr, old, new):
    material_thickness_input.value = str(u.Quantity(material_thickness_input.value, old).to(new).value)
material_thick_unit.on_change('value', update_mat_thick_units)


detector_thick_unit = Select(title="unit", value=length_units[1],
                             options=length_units)


def update_det_thick_units(attr, old, new):
    detector_thickness_input.value = str(u.Quantity(detector_thickness_input.value, old).to(new).value)


detector_thick_unit.on_change('value', update_det_thick_units)
air_thick_unit = Select(title="unit", value=length_units[0],
                        options=length_units)


def update_air_thick_units(attr, old, new):
    air_thickness_input.value = str(u.Quantity(air_thickness_input.value, old).to(new).value)


air_thick_unit.on_change('value', update_air_thick_units)

material_density_unit = Select(title="unit", value=density_units[0],
                               options=density_units)


def update_mat_density_units(attr, old, new):
    material_density_input.value = str(u.Quantity(material_density_input.value, old).to(new).value)


material_density_unit.on_change('value', update_mat_density_units)

detector_density_unit = Select(title="unit", value=density_units[0], options=density_units)


def update_det_density_units(attr, old, new):
    detector_density_input.value = str(u.Quantity(detector_density_input.value, old).to(new).value)


detector_density_unit.on_change('value', update_det_density_units)
air_pressure_unit = Select(title="unit", value=pressure_units[2],
                           options=pressure_units)


def update_air_pressure_units(attr, old, new):
    air_pressure = convert_air_pressure(air_pressure_input.value, old, new)
    air_pressure_input.value = str(air_pressure)


air_pressure_unit.on_change('value', update_air_pressure_units)

air_temp_unit = Select(title="unit", value=temperature_units[1],
                       options=temperature_units)


def update_air_temp_units(attr, old, new):
    air_temperature_input.value = str(u.Quantity(air_temperature_input.value, old).to(new, equivalencies=u.temperature()).value)


air_temp_unit.on_change('value', update_air_temp_units)


def update_material_density(attr, old, new):
    # if the material is changed then update the density
    material_density_input.value = str(get_material_density(material_input.value).value)


material_input.on_change('value', update_material_density)


def update_detector_density(attr, old, new):
    # if the material is changed then update the density
    detector_density_input.value = str(get_material_density(detector_material_input.value).value)


detector_material_input.on_change('value', update_detector_density)

curdoc().add_root(
    layout(
        [
            [plot],
            [checkbox_group, plot_checkbox_group],
            [material_input, material_thickness_input, material_thick_unit, material_density_input, material_density_unit],
            [air_pressure_input, air_pressure_unit, air_thickness_input, air_thick_unit, air_temperature_input, air_temp_unit],
            [detector_material_input, detector_thickness_input, detector_thick_unit, detector_density_input, detector_density_unit],
            #[energy_low, energy_high, energy_step],
            [download_button, Spacer(), update_plot_button],
            [p]
        ],
        sizing_mode="scale_width",
    )
)
curdoc().title = "Roentgen"



"""
"""
import os

import numpy as np
from scipy import interpolate

import astropy.units as u

import roentgen
from roentgen.util import (
    get_atomic_number,
    get_compound_index,
    get_material_density,
    get_material_name,
    get_material_symbol,
    is_an_element,
    is_in_known_compounds,
)

__all__ = ["Material", "MassAttenuationCoefficient", "Stack", "Response"]

_package_directory = roentgen._package_directory
_data_directory = roentgen._data_directory


class Material(object):
    """
    An object which enables the calculation of the x-ray transmission and
    absorption of a material (e.g. an element or a compound/mixture).

    A material may be composed of a single atomic element such as Aluminum ('Al'), or composed of number of elements and/or compounds.

    Parameters
    ----------
    material_str : str or dict
        A string representation of the material which includes an element symbol
        (e.g. Si), an element name (e.g. Silicon), or the name of a compound
        (e.g. cdte, mylar). For all supported elements see :download:`elements.csv <../../roentgen/data/elements.csv>` and for compounds see :download:`compounds_mixtures.csv <../../roentgen/data/compounds_mixtures.csv>`.
        Can also be a dictionary of element and compounds with fractional masses (ex. {"Cu":0.70, "Zn":0.30})
    thickness : `astropy.units.Quantity`
        The thickness of the material
    density : `astropy.units.Quantity`, optional
        The density of the material.
        If not provided, uses default values which can be found in :download:`elements.csv <../../roentgen/data/elements.csv>` for elements or
        in :download:`compounds_mixtures.csv <../../roentgen/data/compounds_mixtures.csv>` for compounds.
        If many materials are present, calculates the weighted density.

    Attributes
    ----------
    symbols : `list`
        A list of material symbol
    material_names : `list`
        A list of material names
    name : `str`
        A name for the material
    fractional_masses : `np.ndarray`
        A normalized array of fractional masses
    mass_attenuation_coefficients : `list`
        A list of `MassAttenuationCoefficient`

    Methods
    -------
    mass_attenuation_coefficient(energy)
        The mass attenuation coefficient for the material at energy

    Examples
    --------
    >>> from roentgen.absorption.material import Material
    >>> import astropy.units as u
    >>> detector = Material('cdte', 500 * u.um)
    >>> thermal_blankets = Material('mylar', 0.5 * u.mm)
    >>> bronze = Material({"Cu": 0.88, "Sn": 0.12}, 1 * u.mm)
    """

    @u.quantity_input
    def __init__(self, material_input, thickness: u.m, density=None):
        self.thickness = thickness
        if isinstance(density, u.Quantity):
            self.density = density
        if isinstance(material_input, str):
            self.list_names = [get_material_name(material_input)]
            self.list_symbols = [get_material_symbol(material_input)]
            self.mass_attenuation_coefficients = [MassAttenuationCoefficient(material_input)]
            self.symbol = self.mass_attenuation_coefficients[0].symbol
            self.name = self.mass_attenuation_coefficients[0].name
            self.fractional_masses = np.ones(1)
            if density is None:
                self.density = get_material_density(material_input)
        elif isinstance(material_input, dict):
            self.list_names = [
                get_material_name(this_str) for this_str in list(material_input.keys())
            ]
            self.list_symbols = [
                get_material_symbol(this_str) for this_str in list(material_input.keys())
            ]
            # normalize the fractional masses
            fractional_masses = np.array(list(material_input.values()))
            self.fractional_masses = fractional_masses / fractional_masses.sum()
            self.name = "".join(f"{this_name}" for this_name in self.list_names)
            self.symbol = "".join(f"{this_symbol}" for this_symbol in self.list_symbols)
            self.mass_attenuation_coefficients = [
                MassAttenuationCoefficient(e) for e in self.list_names
            ]
            if density is None:
                # calculate the average weighted density
                densities = (
                    [
                        get_material_density(this_material).to("kg/m**3").value
                        for this_material in self.list_names
                    ]
                    * u.kg
                    / u.m**3
                )
                self.density = np.average(densities, weights=self.fractional_masses)
        else:
            raise TypeError("Material input must be a string or a dictionary.")

    def __repr__(self):
        """Returns a developer-relevant representation."""
        # at this point, no reason for this to be different than __str__
        return self.__str__()

    def __str__(self):
        """Returns a human-readable user-focused representation."""
        txt = f"Material('{self.name}' thickness={self.thickness} density={self.density.to('kg/m**3'):2.1f})"
        return txt

    def __add__(self, other):
        if isinstance(other, Material):
            return Stack([self, other])
        elif isinstance(other, Stack):
            return Stack([self] + other.materials)
        else:
            raise TypeError(f"Cannot add {self} and {other}")

    @u.quantity_input(energy=u.keV)
    def mass_attenuation_coefficient(self, energy):
        result = np.sum(
            np.vstack(
                [
                    atten.func(energy) * frac_mass
                    for atten, frac_mass in zip(
                        self.mass_attenuation_coefficients, self.fractional_masses
                    )
                ]
            ),
            axis=0,
        )
        if energy.isscalar:
            return result[0]
        else:
            return result

    @u.quantity_input(energy=u.keV)
    def transmission(self, energy):
        """Provide the transmission fraction (0 to 1).

        Parameters
        ----------
        energy : `astropy.units.Quantity`
            An array of energies in keV

        Raises
        ------
        ValueError
            If energy is outside of the interpolation range of 1 keV to 20 MeV.
        """
        coefficients = self.mass_attenuation_coefficient(energy)
        transmission = np.exp(-coefficients * self.density * self.thickness)
        return transmission.value  # remove the dimensionless unit

    @u.quantity_input(energy=u.keV)
    def absorption(self, energy):
        """Provides the absorption fraction (0 to 1).

        Parameters
        ----------
        energy : `astropy.units.Quantity`
            An array of energies in keV.

        Raises
        ------
        ValueError
            If energy is outside of the interpolation range of 1 keV to 20 MeV.
        """
        return 1.0 - self.transmission(energy)

    def linear_attenuation_coefficient(self, energy: u.keV):
        """Provides the linear attenuation coefficient as a function of energy.

        linear coeff = mass coeff * density.

        Parameters
        ----------
        energy : `astropy.units.Quantity`
            An array of energies in keV.

        Raises
        ------
        ValueError
            If energy is outside of the interpolation range of 1 keV to 20 MeV.
        """
        return self.mass_attenuation_coefficient(energy) * self.density


class Stack(object):
    """
    An object which enables the calculation of the x-ray transmission and
    absorption of a stack of materaials.
    This object is created automatically when `Material` objects are added together.

    Parameters
    ----------
    materials : list
        A list of `Material` objects

    Examples
    --------
    >>> from roentgen.absorption.material import Material, Stack
    >>> import astropy.units as u
    >>> detector = Stack([Material('Pt', 5 * u.um), Material('cdte', 500 * u.um)])
    >>> optical_path = Material('mylar', 50 * u.micron) + Material('Al', 1 * u.mm)
    """

    def __init__(self, materials):
        self.materials = materials

    def __add__(self, other):
        if isinstance(other, Material):
            return Stack(self.materials + [other])
        elif isinstance(other, Stack):
            return Stack(self.materials + other.materials)
        else:
            raise TypeError(f"Cannot add {self} and {other}")

    def __repr__(self) -> str:
        """Returns a developer-relevant representation."""
        # at this point, no reason for this to be different than __str__
        return self.__str__()

    def __str__(self) -> str:
        """Returns a human-readable user-focused representation."""
        txt = "Stack(["
        for this_material in self.materials:
            txt += f"{this_material}, "
        txt = f"{txt[:-2]}])"
        return txt

    def transmission(self, energy):
        """Provides the transmission fraction (0 to 1).

        Parameters
        ----------
        energy : `astropy.units.Quantity`
            An array of energies in keV

        Raises
        ------
        ValueError
            If energy is outside of the interpolation range of 1 keV to 20 MeV.
        """
        transmission = np.ones(len(energy), dtype=float)
        for material in self.materials:
            this_transmission = material.transmission(energy)
            transmission *= this_transmission
        return transmission

    def absorption(self, energy):
        """Provides the absorption fraction (0 to 1).

        Parameters
        ----------
        energy : `astropy.units.Quantity`
            An array of energies in keV.

        Raises
        ------
        ValueError
            If energy is outside of the interpolation range of 1 keV to 20 MeV.
        """
        return 1.0 - self.transmission(energy)


class Response(object):
    """
    An object to handle the response of a detector material which includes
    an optical path or filter through which x-rays must first traverse before
    reaching the detector.

    Parameters
    ----------
    optical_path : Stack
        A list of Material objects which make up the optical path.

    detector : Material or None
        A Material which represents the detector material where the x-rays
        are absorbed. If provided with None, than assume a perfectly absorbing
        detector material.

    Examples
    --------
    >>> from roentgen.absorption.material import Material, Response, Stack
    >>> import astropy.units as u
    >>> optical_path = Stack([Material('air', 1 * u.m), Material('Al', 500 * u.mm)])
    >>> resp = Response(optical_path, detector=Material('cdte', 500 * u.um))
    """

    def __init__(self, optical_path, detector):
        # make sure the materials are a list since we iterate over them
        # to calculate the transmission
        if isinstance(optical_path, Stack) or isinstance(optical_path, Material):
            self.optical_path = optical_path
        else:
            raise TypeError("optical_path must be a Stack or Material")

        if isinstance(detector, Material):
            self.detector = detector
        else:
            raise TypeError("detector must be a Material")

    def __repr__(self) -> str:
        """Returns a developer-relevant representation."""
        # at this point, no reason for this to be different than __str__
        return self.__str__()

    def __str__(self) -> str:
        """Returns a human-readable user-focused representation."""
        txt = f"Response(optical_path={self.optical_path} detector={self.detector})"
        return txt

    def response(self, energy):
        """Returns the response as a function of energy which corresponds to the
        transmission through the optical path multiplied by the absorption in
        the detector.

        Parameters
        ----------
        energy : `astropy.units.Quantity`
            An array of energies in keV.

        Raises
        ------
        ValueError
            If energy is outside of the interpolation range of 1 keV to 20 MeV.
        """
        # calculate the transmission
        transmission = np.ones(len(energy), dtype=float)
        detector_absorption = np.ones(len(energy), dtype=float)

        transmission = self.optical_path.transmission(energy)
        detector_absorption = self.detector.absorption(energy)

        return transmission * detector_absorption


class MassAttenuationCoefficient(object):
    """
    The mass attenuation coefficient.

    Parameters
    ----------
    material_str : str
        A string representation of the material which includes an element symbol
        (e.g. Si), an element name (e.g. Silicon), or the name of a compound
        (e.g. cdte, mylar).

    Attributes
    ----------
    data : `astropy.units.Quantity` array
        The mass attenuation data values.
    energy : `astropy.units.Quantity`
        The energy values of the mass attenuation values.
    symbol : `str`
        The material symbol
    name : `str`
        The material name
    func : `lambda func`
        A function which returns the interpolated mass attenuation value at
        any given energy. Energies must be given by an `astropy.units.Quantity`.
        The interpolation range is 1 keV to 20 MeV.
        Going outside that range will result in a ValueError.

    Examples
    --------
    >>> from roentgen.absorption.material import MassAttenuationCoefficient
    >>> mass_atten = MassAttenuationCoefficient('air')
    """

    def __init__(self, material):
        """
        Parameters
        ----------
        material : str
            A string representation of the material which includes an element symbol
            (e.g. Si), an element name (e.g. Silicon), or the name of a compound
            (e.g. cdte, mylar).
        """
        if is_an_element(material):
            atomic_number = get_atomic_number(material)
            datafile_path = os.path.join(
                _data_directory, "elements", "z" + str(atomic_number).zfill(2) + ".csv"
            )
            symbol = roentgen.elements[atomic_number - 1]["symbol"]
            name = roentgen.elements[atomic_number - 1]["name"]
        elif is_in_known_compounds(material):
            compound_index = get_compound_index(material)
            symbol = roentgen.compounds[compound_index]["symbol"]
            name = roentgen.compounds[compound_index]["name"]
            datafile_path = os.path.join(
                _data_directory, "compounds_mixtures", symbol.replace(" ", "_") + ".csv"
            )
        else:
            raise ValueError(f"Element or compound {material} not found.")
        data = np.loadtxt(datafile_path, delimiter=",")
        # find the material in our list
        self.symbol = symbol
        self.name = name
        self.energy = u.Quantity(data[:, 0] * 1000, "keV")
        self.data = u.Quantity(data[:, 1], "cm^2/g")

        self._remove_double_vals_from_data()

        data_energy_kev = np.log10(self.energy.value)
        data_attenuation_coeff = np.log10(self.data.value)
        self._f = interpolate.interp1d(
            data_energy_kev,
            data_attenuation_coeff,
            bounds_error=True,
            assume_sorted=True,
        )
        self.func = lambda x: u.Quantity(10 ** self._f(np.log10(x.to("keV").value)), "cm^2/g")

    def __repr__(self) -> str:
        """Returns a developer-relevant representation."""
        # at this point, no reason for this to be different than __str__
        return self.__str__()

    def __str__(self) -> str:
        """Returns a human-readable user-focused representation."""
        txt = f"MassAttenuationCoefficient('{self.name}')"
        return txt

    def _remove_double_vals_from_data(self):
        """Remove double-values energy values. Edges are represented with
        the same energy index and at the bottom and top value of the edge. This
        must be removed to enable correct interpolation."""
        uniq, count = np.unique(self.energy, return_counts=True)
        duplicates = uniq[count > 1]
        for this_dup in duplicates:
            ind = (self.energy == this_dup).nonzero()
            # shift the first instance of the energy, the bottom of the edge
            self.energy[ind[0][0]] -= 1e-3 * u.eV



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gui',
 'roentgen',
 'roentgen.absorption',
 'roentgen.lines',
 'roentgen.tests',
 'roentgen.util']

package_data = \
{'': ['*'],
 'gui': ['templates/*'],
 'roentgen': ['data/README.rst',
              'data/README.rst',
              'data/README.rst',
              'data/README.rst',
              'data/README.rst',
              'data/README.rst',
              'data/README.rst',
              'data/compounds_mixtures.csv',
              'data/compounds_mixtures.csv',
              'data/compounds_mixtures.csv',
              'data/compounds_mixtures.csv',
              'data/compounds_mixtures.csv',
              'data/compounds_mixtures.csv',
              'data/compounds_mixtures.csv',
              'data/compounds_mixtures/*',
              'data/cxro/*',
              'data/electron_binding_energies.csv',
              'data/electron_binding_energies.csv',
              'data/electron_binding_energies.csv',
              'data/electron_binding_energies.csv',
              'data/electron_binding_energies.csv',
              'data/electron_binding_energies.csv',
              'data/electron_binding_energies.csv',
              'data/elements.csv',
              'data/elements.csv',
              'data/elements.csv',
              'data/elements.csv',
              'data/elements.csv',
              'data/elements.csv',
              'data/elements.csv',
              'data/elements/*',
              'data/emission_energies.csv',
              'data/emission_energies.csv',
              'data/emission_energies.csv',
              'data/emission_energies.csv',
              'data/emission_energies.csv',
              'data/emission_energies.csv',
              'data/emission_energies.csv',
              'data/emission_lines.csv',
              'data/emission_lines.csv',
              'data/emission_lines.csv',
              'data/emission_lines.csv',
              'data/emission_lines.csv',
              'data/emission_lines.csv',
              'data/emission_lines.csv',
              'data/siegbahn_to_iupac.csv',
              'data/siegbahn_to_iupac.csv',
              'data/siegbahn_to_iupac.csv',
              'data/siegbahn_to_iupac.csv',
              'data/siegbahn_to_iupac.csv',
              'data/siegbahn_to_iupac.csv',
              'data/siegbahn_to_iupac.csv']}

install_requires = \
['astropy>=5.1,<6.0', 'numpy>=1.23.3,<2.0.0', 'scipy>=1.9.1,<2.0.0']

setup_kwargs = {
    'name': 'roentgen',
    'version': '2.1.0',
    'description': 'A Python package for the quantitative analysis of the interaction of energetic x-rays with matter. This package is named after one of the discoverers of X-rays, Wilhelm Rontgen.',
    'long_description': '========\nOverview\n========\n\n.. start-badges\n\n.. list-table::\n    :stub-columns: 1\n\n    * - docs\n      - |docs|\n    * - build status\n      - |testing| |codestyle| |coverage|\n    * - package\n      - |version| |downloads| |wheel|\n\n.. |docs| image:: https://readthedocs.org/projects/roentgen/badge/?version=latest\n    :target: https://roentgen.readthedocs.io/en/latest/?badge=latest\n    :alt: Documentation Status\n\n.. |testing| image:: https://github.com/ehsteve/roentgen/actions/workflows/testing.yml/badge.svg\n    :target: https://github.com/ehsteve/roentgen/actions/workflows/testing.yml\n    :alt: Build Status\n\n.. |codestyle| image:: https://github.com/ehsteve/roentgen/actions/workflows/codestyle.yml/badge.svg\n    :target: https://github.com/ehsteve/roentgen/actions/workflows/codestyle.yml\n    :alt: Black linting\n\n.. |coverage| image:: https://codecov.io/gh/ehsteve/roentgen/branch/master/graph/badge.svg?token=feNCnYTjB3\n    :alt: Test coverage on codecov\n    :target: https://codecov.io/gh/ehsteve/roentgen\n\n.. |version| image:: https://img.shields.io/pypi/v/roentgen.svg?style=flat\n    :alt: PyPI Package latest release\n    :target: https://pypi.python.org/pypi/roentgen\n\n.. |downloads| image:: https://img.shields.io/pypi/dm/roentgen.svg?style=flat\n    :alt: PyPI Package monthly downloads\n    :target: https://pypi.python.org/pypi/roentgen\n\n.. |wheel| image:: https://img.shields.io/pypi/wheel/roentgen.svg?style=flat\n    :alt: PyPI Wheel\n    :target: https://pypi.python.org/pypi/roentgen\n\n.. end-badges\n\n.. image:: https://raw.githubusercontent.com/ehsteve/roentgen/main/docs/logo/roentgen_logo.svg\n    :height: 150\n    :width: 150\n\nA Python package for the quantitative analysis of the interaction of energetic x-rays with matter.\nThis package is named after one of the discoverers of X-rays, `Wilhelm Röntgen <https://en.wikipedia.org/wiki/Wilhelm_Röntgen>`_.\n\nInstallation\n============\n\n::\n\n    pip install roentgen\n\nThis project makes use of `Poetry <https://python-poetry.org>`_ for dependency management. To install this project for development, clone the repository and then run the following command inside the package directory\n\n::\n\n    poetry install --with dev,docs,gui\n\n\nDocumentation\n=============\n\nhttp://roentgen.readthedocs.io/en/stable/\n\nGUI\n===\nThis package provides a gui interface to quickly investigate the absorption and transmission of x-rays through different materials.\nIt is based on `bokeh <https://docs.bokeh.org/en/stable/>`_. To run it locally use the following command\n\n::\n\n   bokeh serve --show <roengten_directory>/gui\n\n\nData Sources\n============\nThis package includes on a number of data files which were translated and imported from a few key sources.\nThe package developers would like to thank the following data providers\n\n* The U.S National Institute of Standards and Technology (NIST)\n* The Center for X-ray Optics and Advanced Light Source at the Lawrence Berkeley National Laboratory\n\nFor more information see the `README <roentgen/data/README.rst>`_ in data directory.\n\nContributing\n============\n\nContributions are welcome, and they are greatly appreciated!\nEvery little bit helps, and credit will always be given.\nHave a look at the `great guide <https://docs.sunpy.org/en/latest/dev_guide/contents/newcomers.html>`__ from the `sunpy project <https://sunpy.org>`__ which provides advice for new contributors.\n\nCode of Conduct\n===============\n\nWhen you are interacting with members of this community, you are asked to follow the SunPy `Code of Conduct <https://sunpy.org/coc>`__.\n',
    'author': 'Steven Christe',
    'author_email': 'ehsteve@users.noreply.github.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/ehsteve/roentgen',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.10',
}


setup(**setup_kwargs)
